/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.8 $
 *********************************************************************/

#include "Color.h"
#include <Util.h>

using namespace util;

namespace prapi
{
	double ohtaKanabeData[] = {
		1.0/3.0, 1.0/3.0, 1.0/3.0,
		1, 0, -1,
		-1.0/2.0, 1, -1.0/2.0
	};

	double D65_709_XYZData[] = {
		0.412453, 0.357580, 0.180423,
		0.212671, 0.715160, 0.072169,
		0.019334, 0.119193, 0.950227
	};

	Matrix<double> ColorTransform::ohtaKanabe(3,3,ohtaKanabeData);
	Matrix<double> ColorTransform::D65_709_XYZ(3,3,D65_709_XYZData);

#define LAB_LIMIT				0.008856
#define LAB_F(a)				a = (a > LAB_LIMIT ? pow(a,1.0/3.0) : 7.787*a + 16.0/116.0)
	
	Color<double,3> ColorTransform::XYZToCIELAB(const Color<double,3>& xyzColor,
																							const Color<double,3>& whitePoint)
	{
		double XperXn = xyzColor[0]/whitePoint[0];
		double YperYn = xyzColor[1]/whitePoint[1];
		double ZperZn = xyzColor[2]/whitePoint[2];

		Color<double,3> labColor;

		if (YperYn > LAB_LIMIT)
			labColor[0] = 116*pow(YperYn,1.0/3.0)-16;
		else
			labColor[0] = 903.3*YperYn;

		LAB_F(XperXn);
		LAB_F(YperYn);
		LAB_F(ZperZn);

		labColor[1] = 500*(XperXn - YperYn);
		labColor[2] = 200*(YperYn - ZperZn);

		return labColor;
	}
	
	template <> Color<int,3> ColorTransform::RGBToHSV(const Color<int,3>& rgbColor)
	{
		HSVColor<int> result;

		int r = rgbColor._aComponents[0], g = rgbColor._aComponents[1], b = rgbColor._aComponents[2];
		int &h = result._aComponents[0], &s = result._aComponents[1], &v = result._aComponents[2];
		
		int min = minimum(minimum(r,g),b);
		int max = maximum(maximum(r,g),b);
		v = max; //V

		double delta = max - min;

		if (max != 0)
			s = int(255*(delta / double(max))+0.5); //S
		else
			{
				// r = g = b = 0
				s = 0; //S				
				h = 0; //H undefined -> let's say 0
				return result;
			}

		double tmp=0.0;
		if(delta == 0) h=0;
		else if (r == max)
			tmp = double(g - b) / delta;         // between yellow & magenta
		else if (g == max)
			tmp = 2 + double(b - r) / delta;     // between cyan & yellow
		else
			tmp = 4 + double(r - g) / delta;     // between magenta & cyan

		tmp *= 60;                       // degrees
		if (tmp < 0)
			tmp += 360;

		h = int(tmp+0.5);

		return result;
	}

	template <> Color<int,3> ColorTransform::HSVToRGB(const Color<int,3>& hsvColor)
	{
		RGBColor<int> result;
		int h = hsvColor._aComponents[0], s = hsvColor._aComponents[1], v = hsvColor._aComponents[2];
		int &r = result._aComponents[0], &g = result._aComponents[1], &b = result._aComponents[2];

		int i, p, q, t;;
		double f;

		if(s == 0)
			{
				// achromatic (gray)
				r = g = b = v;
				return result;
			}

		double tmpH = double(h)/60; // sector 0 to 5
		double tmpS = double(s)/255;
		i = int(tmpH);
		f = tmpH - i;                      // factorial part of h
		p = int(v * (1.0 - tmpS) + 0.5);
		q = int(v * (1.0 - tmpS * f) + 0.5);
		t = int(v * (1.0 - tmpS * (1 - f)) + 0.5);

		switch(i)
			{
			case 0:
				r = v;
				g = t;
				b = p;
				break;
			case 1:
				r = q;
				g = v;
				b = p;
				break;
			case 2:
				r = p;
				g = v;
				b = t;
				break;
			case 3:
				r = p;
				g = q;
				b = v;
				break;
			case 4:
				r = t;
				g = p;
				b = v;
				break;
			default:                // case 5:
				r = v;
				g = p;
				b = q;
				break;
			}
		return result;
	}
}
