/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.19 $
 *********************************************************************/

#ifndef _COLOR_H
#define _COLOR_H

#include <iostream>
#include <string.h>
#include <Matrix.h>
#include <math.h>
#include <SmartPtr.h>

namespace prapi
{
	template <class T, int comps> class Color;
	
	/**
	 * Contains methods for color space transforms that cannot be
	 * implemeted without some external knowledge. Transformations that
	 * do not need additional information (RGB<->HSV) are implemented
	 * simply as typecasts. This class lso contains static matrices for
	 * color space transforms.
	 **/
	class ColorTransform
	{
	public:
		/**
		 * Convert a color in XYZ color space into the CIELAB space.
		 *
		 * @param xyzColor the color to be converted to CIELAB
		 * @param whitePoint the XYZ coordinates of the white point of the
		 * imaging equipment. Make sure this value has no zero components.
		 **/
		static Color<double,3> XYZToCIELAB(const Color<double,3>& xyzColor,
																			 const Color<double,3>& whitePoint);


		/**
		 * Convert a color from the RGB space to HSV. If the type of the
		 * color is float or double, values must be in the range [0,1]. If
		 * the type is int, values must be in the range [0,255]. For other
		 * types, the behavior is undefined.
		 *
		 * @param rgbColor the color to be converted
		 * @return the same color in the HSV space
		 **/
		template <class T> static Color<T,3> RGBToHSV(const Color<T,3>& rgbColor);

		/**
		 * Convert a color from the HSV space to RGB. For integer-valued
		 * HSV colors, the saturation channel must be in the range
		 * [0,255]. Preferably, the value channel should also be in this
		 * range. The hue channel must be in [0,360) independent of the
		 * content type. For floating point content types, saturation must
		 * and value should be in the range [0,1].
		 *
		 * @param hsvColor the color to be converted
		 * @return the same color in the RGB space
		 **/
		template <class T> static Color<T,3> HSVToRGB(const Color<T,3>& hsvColor);
		
		/**
		 * Normalize a color. A normalized color is created by dividing
		 * color channels by the sum of all channels. Typically, this
		 * operation is used to get chromaticity information.
		 **/
		template <class T, int c> static Color<double,c> normalize(const Color<T,c>& clr);
		
		/**
		 * A 3-by-3 matrix for transforming a color from the RGB color
		 * model to the one defined by Ohta and Kanabe.
		 **/
		static util::Matrix<double> ohtaKanabe;

		/**
		 * The "standard" matrix for RGB to XYZ conversion. Assumes ITU-R
		 * BT.709 camera primaries and D65 white point.
		 **/
		static util::Matrix<double> D65_709_XYZ;
	};


	template <class T, int c> Color<double,c> ColorTransform::normalize(const Color<T,c>& clr)
	{
		T sum = 0;
		for (int i=c;i--;)
			sum += clr[i];
		if (sum == 0)
			return Color<double,c>(0);
		return Color<double,c>(clr)/sum;
	}

	template <> Color<int,3> ColorTransform::RGBToHSV(const Color<int,3>& rgbColor);

	template <class T> Color<T,3> ColorTransform::RGBToHSV(const Color<T,3>& rgbColor)
	{
		Color<T,3> result;

		T r = rgbColor._aComponents[0], g = rgbColor._aComponents[1], b = rgbColor._aComponents[2];
		T &h = result._aComponents[0], &s = result._aComponents[1], &v = result._aComponents[2];
		
		T min = minimum(minimum(r,g),b);
		T max = maximum(maximum(r,g),b);
		v = max; //V

		T delta = max - min;

		if (max != 0)
			s = delta / max; //S
		else
			{
				// r = g = b = 0
				s = 0; //S				
				h = 0; //H undefined -> let's say 0
				return result;
			}

		if(delta == 0)h=0;
		else if (r == max)
			h = (g - b) / delta;         // between yellow & magenta
		else if (g == max)
			h = 2 + (b - r) / delta;     // between cyan & yellow
		else
			h = 4 + (r - g) / delta;     // between magenta & cyan

		h *= 60;                       // degrees
		if (h < 0)
			h += 360;

		return result;
	}

	template <> Color<int,3> ColorTransform::HSVToRGB(const Color<int,3>& hsvColor);

	template <class T> Color<T,3> ColorTransform::HSVToRGB(const Color<T,3>& hsvColor)
	{
		Color<T,3> result;
		T h = hsvColor._aComponents[0], s = hsvColor._aComponents[1], v = hsvColor._aComponents[2];
		T &r = result._aComponents[0], &g = result._aComponents[1], &b = result._aComponents[2];

		int i;
		T f, p, q, t;

		if(s == 0)
			{
				// achromatic (gray)
				r = g = b = v;
				return result;
			}

		h /= 60;                        // sector 0 to 5
		i = (int)floor(h);
		f = h - i;                      // factorial part of h
		p = v * (1 - s);
		q = v * (1 - s * f);
		t = v * (1 - s * (1 - f));

		switch(i)
			{
			case 0:
				r = v;
				g = t;
				b = p;
				break;
			case 1:
				r = q;
				g = v;
				b = p;
				break;
			case 2:
				r = p;
				g = v;
				b = t;
				break;
			case 3:
				r = p;
				g = q;
				b = v;
				break;
			case 4:
				r = t;
				g = p;
				b = v;
				break;
			default:                // case 5:
				r = v;
				g = p;
				b = q;
				break;
			}
		return result;
	}
	
	/**
	 * A universal representation of a color. A color may have any
	 * number of components that can be stored as any data types. It is
	 * advisable not to use class types as color components because
	 * color copies are implemented as raw memory transfers.
	 * Constructors are thus not called.
	 **/
	template <class T, int comps> class Color
	{
	public:
		/**
		 * The default constructor. Initializes the component array by
		 * filling it with zero bytes. Using floating point component
		 * values may result in weird initial values and you may want to
		 * use the Color(value) constructor instead.
		 **/
		Color() { memset(_aComponents,0,sizeof(T)*comps); }
		/**
		 * Construct a new color whose components are reset to the given
		 * value.
		 * @param value the initial value for all of the color components
		 **/
		Color(T value) { for (int i=0;i<comps;i++) _aComponents[i] = value; }
		/**
		 * The copy constructor.
		 **/
		Color(const Color& other) { memcpy(_aComponents,other._aComponents,sizeof(T)*comps); }

		/**
		 * Create a typecasted copy of another color.
		 **/
		template <class U> Color(const Color<U,comps>& other);

		/**
		 * Get a color component. For example, in an RGB color, index 0
		 * represents the green channel and index 2 represents the blue
		 * channel.
		 * @param index the channel index
		 **/
		T getComponent(int index) { return _aComponents[index]; }
		/**
		 * Set the value of a color channel.
		 * @param index the channel index
		 * @param value the new value for this channel
		 **/
		void setComponent(int index, T value) { _aComponents[index] = value; }
		/**
		 * Get the color components as an array.
		 * @return a pointer to the internal color component array
		 **/
		T* getComponents(void) { return _aComponents; }
		/**
		 * Set the color components all at once.
		 * @param array an array with an entry for each color channel.
		 **/
		void setComponents(T* array) { memcpy(_aComponents,array,sizeof(T)*comps); }

		/**
		 * Copy a color.
		 **/
		Color& operator= (const Color& other) { memcpy(_aComponents,other._aComponents,sizeof(T)*comps); return *this; }
		/**
		 * Set all color channels to the same value.
		 **/
		Color& operator= (T value) { for (int i=0;i<comps;i++) _aComponents[i] = value; return *this; }

		/**
		 * Compare two colors. Colors are equal if all of their
		 * corresponding components are equal.
		 **/
		template <class U, int nc> friend bool operator== (const Color<U,nc>& clr1, const Color<U,nc>& clr2);
		/**
		 * Compare two colors. Colors are different if any of their
		 * corresponding components are different.
		 **/
		template <class U, int nc> friend bool operator!= (const Color<U,nc>& clr1, const Color<U,nc>& clr2);

		/**
		 * Get a given color channel.
		 * @see getComponent(int)
		 **/
		T& operator[] (int index) { return _aComponents[index]; }
		/**
		 * Get a given color channel.
		 * @see getComponent(int)
		 **/
		T operator[] (int index) const { return _aComponents[index]; }

		/**
		 * Add the corresponding color channels together.
		 **/
		void operator+= (Color other);
		/**
		 * Subract the corresponding color channels from each other.
		 **/
		void operator-= (Color other);
		/**
		 * Multiply corresponding color channels by each other.
		 **/
		void operator*= (Color other);
		/**
		 * Divide corresponding color channels by each other.
		 **/
		void operator/= (Color other);

		/**
		 * Multiply all channels by a value.
		 **/
		void operator*= (double value);
		/**
		 * Divide all channels by a value.
		 **/
		void operator/= (double value);
		/**
		 * Add a value to all color channels.
		 **/
		void operator+= (double value);
		/**
		 * Subtract a value from all color channels.
		 **/
		void operator-= (double value);

		/**
		 * Multiply all channes by a value.
		 **/
		template <class U, int nc> friend Color<U,nc> operator* (const Color<U,nc>& clr, double value);
		/**
		 * Divide all channes by a value.
		 **/
		template <class U, int nc> friend Color<U,nc> operator/ (const Color<U,nc>& clr, double value);
		/**
		 * Add a value to all channes.
		 **/
		template <class U, int nc> friend Color<U,nc> operator+ (const Color<U,nc>& clr, double value);
		/**
		 * Subtract a value from all channes.
		 **/
		template <class U, int nc> friend Color<U,nc> operator- (const Color<U,nc>& clr, double value);

		/**
		 * Sum corresponding channels together.
		 **/
		template <class U, int nc> friend Color<U,nc> operator+ (const Color<U,nc>& clr1, const Color<U,nc>& clr2);
		/**
		 * Subtract corresponding channels from each other.
		 **/
		template <class U, int nc> friend Color<U,nc> operator- (const Color<U,nc>& clr1, const Color<U,nc>& clr2);
		/**
		 * Multiply corresponding channels by each other.
		 **/
		template <class U, int nc> friend Color<U,nc> operator* (const Color<U,nc>& clr1, const Color<U,nc>& clr2);
		/**
		 * Divide corresponding channels by each other.
		 **/
		template <class U, int nc> friend Color<U,nc> operator/ (const Color<U,nc>& clr1, const Color<U,nc>& clr2);

		/**
		 * Cast a color to a different type of color.
		 **/
		template <class U> operator Color<U,comps>();
		/**
		 * Make a typecasted copy of a color.
		 **/
		template <class U> Color& operator= (const Color<U,comps>& other);

		/**
		 * Write to a stream.
		 **/
		template <class U, int nc> friend std::ostream& operator<< (std::ostream& sout, const Color<U,nc>& c);
		/**
		 * Read from a stream.
		 **/
		template <class U, int nc> friend std::istream& operator>> (std::istream& sin, Color<U,nc>& c);

		/**
		 * Transform this a color by a comps-by-comps transformation matrix.
		 **/
		template <class U, class V, int nc> friend Color<double,nc> operator* (const util::Matrix<V>& mat,
																																					 const Color<U,nc>& clr) throw (util::MatrixException&);

		friend class ColorTransform;
	protected:
		/**
		 * The component array.
		 **/
		T _aComponents[comps];
	};

	template <class T,int comps>
	template <class U> Color<T,comps>::Color(const Color<U,comps>& other)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] = (T)other[i];
	}

	template <class T,int comps>
	template <class U> Color<T,comps>::operator Color<U,comps>()
	{
		Color<U,comps> result(0);
		for (int i=0;i<comps;i++)
			result[i] = (U)_aComponents[i];
		return result;
	}

	template <class T,int comps>
	template <class U> Color<T,comps>& Color<T,comps>::operator= (const Color<U,comps>& other)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] = (T)other[i];
		return *this;
	}

	template <class T, int comps> std::ostream& operator<< (std::ostream& sout, const Color<T,comps>& c)
	{
		sout << "<color>";
		for (int i=0;i<comps;i++)
			{
				if (i) sout << " ";
				util::Util::writeXMLItem(sout,c._aComponents[i]);
			}
		sout << "</color>";
		return sout;
	}

	template <class T, int comps> std::istream& operator>> (std::istream& sin, Color<T,comps>& c)
	{
		using namespace util::xml;

		XMLParser parser;
		sin >> ws;
		
		util::SmartPtr<Node> node(parser.readFragment(sin));
		if (node.get() && node->getNodeType() == Node::ELEMENT_NODE)
			{
				Element* element = (Element*)node.get();
				if (element->tagType != Element::TAG_OPENING || element->tagName != "color")
					throw util::io::IOException("operator>> (istream&, Color<T,comps>&): Expecting <color>.");
				Node *text = element->getChildNode("color.text");
				if (!text || text->getNodeType() != Node::TEXT_NODE)
					throw util::io::IOException("operator>> (istream&, Color<T,comps>&): Color does not contain data.");
				util::List<T> components(util::String::parseList<T>(((Text*)text)->contents," \t\n"));
				if (components.getLength() != comps)
					throw util::io::IOException("operator>> (istream&, Color<T,comps>&): Wrong number of color components.");
				for (int i=0;i<comps;i++)
					c._aComponents[i] = components[i];
			}
		else
			throw util::io::IOException("operator>> (istream&, Color<T,comps>&): The input stream does not contain a color.");
		
		return sin;
	}

	template <class T, int comps> void Color<T,comps>::operator+= (Color<T,comps> other)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] += other._aComponents[i];
	}
	
	template <class T, int comps> void Color<T,comps>::operator-= (Color<T,comps> other)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] -= other._aComponents[i];
	}

	template <class T, int comps> void Color<T,comps>::operator*= (Color<T,comps> other)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] *= other._aComponents[i];
	}

	template <class T, int comps> void Color<T,comps>::operator/= (Color<T,comps> other)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] /= other._aComponents[i];
	}

	template <class T, int comps> void Color<T,comps>::operator*= (double value)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] = _aComponents[i] * value;
	}
	
	template <class T, int comps> void Color<T,comps>::operator/= (double value)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] = _aComponents[i] / value;
	}

	template <class T, int comps> void Color<T,comps>::operator+= (double value)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] = _aComponents[i] + value;
	}

	template <class T, int comps> void Color<T,comps>::operator-= (double value)
	{
		for (int i=0;i<comps;i++)
			_aComponents[i] = _aComponents[i] - value;
	}
	
	template <class T, int comps> Color<T,comps> operator* (const Color<T,comps>& clr, double value)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = T(value * clr._aComponents[i]);
		return result;
	}

	template <class T, int comps> Color<T,comps> operator/ (const Color<T,comps>& clr, double value)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = T(clr._aComponents[i] / value);
		return result;
	}

	template <class T, int comps> Color<T,comps> operator+ (const Color<T,comps>& clr, double value)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = T(value + clr._aComponents[i]);
		return result;
	}

	template <class T, int comps> Color<T,comps> operator- (const Color<T,comps>& clr, double value)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = T(clr._aComponents[i] - value);
		return result;
	}

	template <class T, int comps> Color<T,comps> operator+ (const Color<T,comps>& clr1, const Color<T,comps>& clr2)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = clr1._aComponents[i] + clr2._aComponents[i];
		return result;
	}
		
	template <class T, int comps> Color<T,comps> operator- (const Color<T,comps>& clr1, const Color<T,comps>& clr2)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = clr1._aComponents[i] - clr2._aComponents[i];
		return result;
	}

	template <class T, int comps> Color<T,comps> operator* (const Color<T,comps>& clr1, const Color<T,comps>& clr2)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = clr1._aComponents[i] * clr2._aComponents[i];
		return result;
	}
		
	template <class T, int comps> Color<T,comps> operator/ (const Color<T,comps>& clr1, const Color<T,comps>& clr2)
	{
		Color<T,comps> result;
		for (int i=0;i<comps;i++)
			result._aComponents[i] = clr1._aComponents[i] / clr2._aComponents[i];
		return result;
	}
		
	template <class T, int comps>	bool operator== (const Color<T,comps>& clr1, const Color<T,comps>& clr2)
	{
		for (int i=0;i<comps;i++)
			if (clr1[i] != clr2[i])
				return false;
		return true;
	}	

	template <class T, int comps>	bool operator!= (const Color<T,comps>& clr1, const Color<T,comps>& clr2)
	{
		for (int i=0;i<comps;i++)
			if (clr1[i] == clr2[i])
				return true;
		return false;
	}	

	template <class T, class U, int comps> Color<double,comps> operator* (const util::Matrix<U>& mat,
																																				const Color<T,comps>& clr) throw (util::MatrixException&)
	{
		if (mat.getColumns() != comps || mat.getRows() != comps)
			throw util::MatrixException("operator*(Matrix, Color): Matrix dimensions do not match the color");

		Color<double,comps> result(0);
		const U* data = mat.getData();
		for (int i=0,index=0;i<comps;i++)
			for (int j=0;j<comps;j++,index++)
				result[i] += clr[j]*data[index];

		return result;
	}

	template <class T> class HSVColor; //A forward declaration for RGBColor
	
	/**
	 * A convenience class for the RGB color space. RGB colors are
	 * represented as three component values. In fact, RGBColor can be
	 * used in with any three-dimensional color space. The default
	 * component class is <i>int</i>.
	 **/
	template <class T=int> class RGBColor : public Color<T,3>
	{
	public:
		friend class HSVColor<T>;

		/**
		 * Create a new RGB color with all channels set to <i>clr</i>.
		 **/
		RGBColor(T clr = 0) : Color<T,3>(clr) {}
		/**
		 * Create a new RGB color with color channels set as indicated.
		 **/
		RGBColor(T r, T g, T b);
		/**
		 * Create a copy of an RGB color of any type.
		 **/
		template <class U> RGBColor(const Color<U,3>& other) : Color<T,3>(other) {}
		/**
		 * The copy constructor.
		 **/
		RGBColor(const Color<T,3>& other) : Color<T,3>(other) {}
		/**
		 * Transform a color in the HSV color space to the RGB color
		 * space.
		 **/
		RGBColor(const HSVColor<T>& other) : Color<T,3>(ColorTransform::HSVToRGB(other)) {}

		/**
		 * Get the value of the red color channel.
		 **/
		T getRed() const { return _aComponents[0]; }
		/**
		 * Get the value of the green color channel.
		 **/
		T getGreen() const { return _aComponents[1]; }
		/**
		 * Get the value of the blue color channel.
		 **/
		T getBlue() const { return _aComponents[2]; }

		/**
		 * Set the red color channel.
		 **/
		void setRed(T value) { _aComponents[0] = value; }
		/**
		 * Set the green color channel.
		 **/
		void setGreen(T value) { _aComponents[1] = value; }
		/**
		 * Set the blue color channel.
		 **/
		void setBlue(T value) { _aComponents[2] = value; }

		/**
		 * Convert an RGB color to an integer. The result is the intensity
		 * of this color, i.e. the average of R, G, and B.
		 **/
		operator int() const { return int((_aComponents[0]+_aComponents[1]+_aComponents[2])/3 + 0.5); }
		//operator int() const { return (299*_aComponents[0]+587*_aComponents[1]+114*_aComponents[2]+500)/1000; }
		/**
		 * Convert an RGB color to a float. The result is the intensity of
		 * this color, i.e. the average of R, G, and B.
		 **/
		operator float() const { return (_aComponents[0]+_aComponents[1]+_aComponents[2])/3.0f; }
		/**
		 * Convert an RGB color to a double. The result is the intensity
		 * of this color, i.e. the average of R, G, and B.
		 **/
		operator double() const { return (_aComponents[0]+_aComponents[1]+_aComponents[2])/3.0; }

		/**
		 * Convert an RGB color to HSV. This method works correctly only
		 * when float or double is used as the content type and RGB color
		 * values are in the range [0,1], or when the content type is int,
		 * and coordinates are in the range [0,255].
		 **/
		operator HSVColor<T>() const { return RGBColor<T>(ColorTransform::RGBToHSV(*this)); }

		/**
		 * Copy a color of any content type.
		 **/
		template <class U> RGBColor& operator= (const Color<U,3>& other) { Color<T,3>::operator=(other); return *this; }
		/**
		 * Copy a color.
		 **/
		RGBColor& operator= (const Color<T,3>& other) { Color<T,3>::operator=(other); return *this; }

		/**
		 * Set all color channels to <i>value</i>.
		 **/
		RGBColor& operator= (T value) { Color<T,3>::operator=(value); return *this; }

		/**
		 * Get a normalized version of this color. Each color channel is
		 * divided by the sum of all channels.
		 **/
		RGBColor<double> getNormalizedColor(void) const;
	};

	typedef RGBColor<int> IntegerRGBColor;
	typedef RGBColor<float> FloatRGBColor;
	typedef RGBColor<double> DoubleRGBColor;
		
	template <class T> RGBColor<T>::RGBColor<T>(T r,T g,T b)
	{
		_aComponents[0] = r;
		_aComponents[1] = g;
		_aComponents[2] = b;
	}

	template <class T> RGBColor<double> RGBColor<T>::getNormalizedColor(void) const
	{
		RGBColor<double> result;
		double intensity = double(_aComponents[0] + _aComponents[1] + _aComponents[2]);
		if(intensity !=0.0)
			{
				result[0] = double(_aComponents[0]) / intensity;
				result[1] = double(_aComponents[1]) / intensity;
				result[2] = double(_aComponents[2]) / intensity;
			}
		else
			{
				result[0]=0.0;
				result[1]=0.0;
				result[2]=0.0;
			}
		
		return result;
	}


	/**
	 * A class for the HSV (hue, saturation, value) color space. The H
	 * channel varies in the range [0,360). S channel is in the range
	 * [0,1] for data types float and double. For int, the range is
	 * [0,255]. The value of V can be just about anything.
	 **/
	template <class T=int> class HSVColor : public Color<T,3>
	{
	public:
		friend class RGBColor<T>;

		/**
		 * Create a new HSV color with all channels set to <i>clr</i>.
		 **/
		HSVColor(T clr = 0) : Color<T,3>(clr) {}
		/**
		 * Create a new HSV color with color channels set as indicated.
		 **/
		HSVColor(T r, T g, T b);

		/**
		 * Copy any three-component color.
		 **/
		template <class U> HSVColor(const Color<U,3>& other) : Color<T,3>(other) {}
		/**
		 * Copy any three-component color with the same content type.
		 **/
		HSVColor(const Color<T,3>& other) : Color<T,3>(other) {}
		/**
		 * Convert an RGB color to HSV.
		 *
		 * @see #operator RGBColor<T>
		 **/
		HSVColor(const RGBColor<T>& other) : Color<T,3>(ColorTransform::RGBToHSV(other)) {}

		/**
		 * Get the value of the 'hue' channel.
		 **/
		T getHue() const { return _aComponents[0]; }
		/**
		 * Get the value of the 'saturation' channel.
		 **/
		T getSaturation() const { return _aComponents[1]; }
		/**
		 * Get the value of the 'value' channel.
		 **/
		T getValue() const { return _aComponents[2]; }

		/**
		 * Set hue.
		 **/
		void setHue(T value) { _aComponents[0] = value; }
		/**
		 * Set saturation.
		 **/
		void setSaturation(T value) { _aComponents[1] = value; }
		/**
		 * Set value.
		 **/
		void setValue(T value) { _aComponents[2] = value; }

		/**
		 * Get the intensity (V) channel as an integer.
		 **/
		operator int() const { return int(_aComponents[0]); }
		/**
		 * Get the intensity (V) channel as a float.
		 **/
		operator float() const { return float(_aComponents[0]); }
		/**
		 * Get the intensity (V) channel as a double.
		 **/
		operator double() const { return double(_aComponents[0]); }

		/**
		 * Convert this color to the RGB color space. For integer-valued
		 * HSV colors, the saturation channel must be in the range
		 * [0,255]. Preferably, the value channel should also be in this
		 * range. The hue channel must be in [0,360) independent of the
		 * content type. For floating point content types, saturation must
		 * and value should be in the range [0,1].
		 **/
		operator RGBColor<T>() const { return HSVColor<T>(ColorTransform::HSVToRGB(*this)); }

		/**
		 * Copy any three-component color.
		 **/
		template <class U> HSVColor& operator= (const Color<U,3>& other) { Color<T,3>::operator=(other); return *this; }
		/**
		 * Copy a three-component color of the same content type.
		 **/
		HSVColor& operator= (const Color<T,3>& other) { Color<T,3>::operator=(other); return *this; }
	};
}

#endif
