/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.9 $
 *********************************************************************/

#ifndef _COLORVIFFCODEC_H
#define _COLORVIFFCODEC_H

#include "ViffCodec.h"
#include "Color.h"
#include <io/IO.h>

namespace prapi
{
	/**
	 * <pre>
	 * This enumeration is for Viff images to decide which kind of image is.
	 * The explanation of the prefixes:
	 *
	 *  ntsc     National Television Systems Committee
   *  cie      Commission Internationale de L'Eclairage
   *  ucs      Universal Chromaticity Scale
	 *  generic  Color image but not conforming to any standard 
	 *
	 * The explanation of the colour spaces:
	 *
	 *  NONE     used when no color space defined
	 *  RGB      red band, green band, blue band
   *  CMY      cyan band, magenta band, yellow band
   *  YIQ      luminance, I and Q represent chrominance
   *  HSV      hue, saturation, value
   *  IHS      intensity, hue, saturation
	 *  HLS 
   *  XYZ     
   *  UVW  
   *  SOW  
   *  Lab
   *  Luv
	 *  GENERIC  the color space is user defined
	 * </pre>
	 **/
	enum ColorSpaceType { NONE, ntscRGB, ntscCMY, ntscYIQ, HSV, HLS, IHS, cieRGB, cieXYZ, cieUVW,
												cieucsUVW, cieucsSOW, cieucsLab, cieucsLuv, GENERIC, genericRGB };
	/**
	 * A codec for all kinds Color Viff images.
	 **/
	template <class T,int comps> class ColorViffCodec : public ViffCodec<Color<T,comps> >
	{
	public:
		/**
		 * Constructor of the ColorViffCodec.
		 *
		 * @param spaceType The type of Color space used in image.
		 **/
		ColorViffCodec(ColorSpaceType colorSpaceType=genericRGB) : _ColorSpaceType(colorSpaceType){}

		void decodeMatrix(std::istream& in, util::Matrix<Color<T,comps> >& mat) throw (util::MatrixException&, util::io::IOException&);
		void encodeMatrix(std::ostream& out, const util::Matrix<Color<T,comps> >& mat) throw (util::MatrixException&, util::io::IOException&);

		/**
		 * Get the whole channel out of image mat.
		 *
		 * @param c The channel wanted.
		 * @param mat The matrix where the channel is wanted to take.
		 **/
		util::Matrix<T> getChannel(int c, util::Matrix<Color<T,comps> >& mat) const;

	private:
		/**
		 * The type of ColorSpace used in image.
		 **/
		ColorSpaceType _ColorSpaceType;
	};

	template <class T, int comps> void ColorViffCodec<T,comps>::decodeMatrix(std::istream& in, util::Matrix<Color<T, comps> >& mat)
		throw (util::MatrixException&, util::io::IOException&)
	{
		int dataStorageType = 0;
		int numberOfBands = 0;
		decodeImageHeader(in,mat,numberOfBands,dataStorageType);
	
		if(numberOfBands != comps)
			throw MatrixException("ViffCoedec<T,comps>::decodeMatrix(istream&, Matrix<Color<T, comps> >&): NumberOfBands differ");

		// take the data out from the matrix
		int pixels = mat.getRows()*mat.getColumns();
		
		switch(dataStorageType)
			{
			case 0: //short matrix
				{
					int size = pixels*sizeof(short);
					unsigned char* data = new unsigned char[size];
					for(int n=0;n<numberOfBands;n++)
						{
							Color<T,comps>* matData = mat.getData();
							inRead(in,(char*)data,size);
							short* buffer = (short*)data;
							for(int i=0;i<pixels;i++,buffer++,matData++)
								matData->setComponent(n, T(*buffer));
						}
					delete[] data;
				}
				break;
			case 1: /// unsigned char matrix
				{
					int size = pixels*sizeof(unsigned char);
					unsigned char* data = new unsigned char[size];
					for(int n=0;n<numberOfBands;n++)
						{
							Color<T,comps>* matData = mat.getData();
							inRead(in,(char*)data,size);
							unsigned char* buffer = data;
							for(int i=0;i<pixels;i++,buffer++,matData++)
								matData->setComponent(n, T(*buffer));
						}
					delete[] data;
				}
				break;
			case 2: /// short matrix
				{
					int size = pixels*sizeof(short);
					unsigned char* data = new unsigned char[size];
					for(int n=0;n<numberOfBands;n++)
						{
							Color<T,comps>* matData = mat.getData();
							inRead(in,(char*)data,size);
							short* buffer = (short*)data;
							for(int i=0;i<pixels;i++,buffer++,matData++)
								matData->setComponent(n, T(*buffer));
							
						}
					delete[] data;
				}
				break;
			case 4: /// int matrix
				{
					int size = pixels*sizeof(int);
					unsigned char* data = new unsigned char[size];
					for(int n=0;n<numberOfBands;n++)
						{
							Color<T,comps>* matData = mat.getData();
							inRead(in,(char*)data,size);
							int* buffer = (int*)data;
							for(int i=0;i<pixels;i++,buffer++,matData++)
								matData->setComponent(n, T(*buffer));
						}
					delete[] data;
				}
				break;
			case 5: /// float matrix
				{
					int size = pixels*sizeof(float);
					unsigned char* data = new unsigned char[size];
					for(int n=0;n<numberOfBands;n++)
						{
							Color<T,comps>* matData = mat.getData();
							inRead(in,(char*)data,size);
							float* buffer = (float*)data;
							for(int i=0;i<pixels;i++,buffer++,matData++)
								matData->setComponent(n, T(*buffer));
						}
					delete[] data;
				}
				break;
			case 9: /// double matrix
				{
						int size = pixels*sizeof(double);
						unsigned char* data = new unsigned char[size];
						for(int n=0;n<numberOfBands;n++)
							{
								Color<T,comps>* matData = mat.getData();
								inRead(in,(char*)data,size);
								double* buffer = (double*)data;
								for(int i=0;i<pixels;i++,buffer++,matData++)
									matData->setComponent(n, T(*buffer));
							}
						delete[] data;
					}
				break;
			default:
				throw MatrixException("ViffCoedec<T, comps>::decodeMatrix(istream&, Matrix<Color<T, comps> >&): Matrix dataStrorage type not supported.");
			}//switch
	}

	template <class T, int comps> void ColorViffCodec<T,comps>::encodeMatrix(std::ostream& out, const util::Matrix<Color<T,comps> >& mat)
		throw (util::MatrixException&, util::io::IOException&)
	{
		// initialize VIFF header
		xvimage *image = new xvimage;
		if(image == NULL)
			throw IOException("ViffCodec<T>::encodeMatrix(ostream&, Matrix<T>&):Cannot allocate memory for VIFF header.");
		
		image->identifier = XV_FILE_MAGIC_NUM;
		image->file_type = XV_FILE_TYPE_XVIFF;
		image->release = XV_IMAGE_REL_NUM;
		image->version = XV_IMAGE_VER_NUM;
		image->machine_dep = VFF_DEP_BIGENDIAN;
		strcpy(image->comment, " ");
		image->row_size = mat.getColumns();
		image->col_size = mat.getRows();
		image->subrow_size = 0;
		image->startx = VFF_NOTSUB;
		image->starty = VFF_NOTSUB;
		image->pixsizx = 1.0;
		image->pixsizy = 1.0;
		image->location_type = VFF_LOC_IMPLICIT;
		image->location_dim = 0;
		image->num_of_images = 1;
		image->num_data_bands = comps;
		T dataType;
		std::string className(Util::getClassName(dataType));
		if(className=="unsigned char")image->data_storage_type = VFF_TYP_1_BYTE;    
		else if(className=="short")image->data_storage_type = VFF_TYP_2_BYTE;    
    else if(className=="int")image->data_storage_type = VFF_TYP_4_BYTE;    
		else if(className=="float")image->data_storage_type = VFF_TYP_FLOAT;
		//else if(className=="complex")image->data_storage_type = VFF_TYP_COMPLEX;    
    else if(className=="double")image->data_storage_type = VFF_TYP_DOUBLE;    
		else image->data_storage_type = VFF_TYP_1_BYTE;
		image->data_encode_scheme = VFF_DES_RAW;
		image->map_scheme = VFF_MS_NONE;
		image->map_storage_type = VFF_MAPTYP_NONE;
		image->map_row_size = 0;
		image->map_col_size = 0;
		image->map_subrow_size = 0;
		image->map_enable = VFF_MAP_OPTIONAL;
		image->maps_per_cycle = 0;
		image->ispare1 = 0;
		image->ispare2 = 0;
		image->fspare1 = 0;
		image->fspare2 = 0;
		image->maps = NULL;
		image->location = NULL;
		image->imagedata = NULL;
		image->color_space_model = (unsigned long)_ColorSpaceType;
		if(int(_ColorSpaceType) == 15)
			image->data_storage_type = VFF_TYP_1_BYTE;
						
		// write VIFF header to output stream
		outWrite(out,(char *)image, sizeof(xvimage));

		// this is therefor that genericRGB can be writen only
		// at the type VFF_TYP_1_BYTE which mean unsigned char.
		if(int(_ColorSpaceType) == 15)
			{
				int size = mat.getRows()*mat.getColumns()*sizeof(unsigned char);
				// write image data layer by layer
				for(int i=0;i<comps;i++)
					{
						util::Matrix<unsigned char> test(getChannel(i,mat));
						unsigned char* data = test.getData();
						outWrite(out,(char*)data,size);
					}
			}
		else
			{
				int size = mat.getRows()*mat.getColumns()*sizeof(T);
				// write image data layer by layer
				for(int i=0;i<comps;i++)
					outWrite(out,(char*)getChannel(i,mat).getData(),size);
			}
		
		delete image;
	}

	template <class T,int comps> util::Matrix<T> ColorViffCodec<T,comps>::getChannel(int c,util::Matrix<Color<T,comps> >& mat) const
	{
		int rows = mat.getRows();
		int cols = mat.getColumns();
		util::Matrix<T> result(rows,cols);
		const Color<T,comps>* data = mat.getData();
		T* rData = result.getData();
		for (int i=0;i<rows*cols;i++)
			rData[i] = data[i][c];
		return result;
	}
	

	
} // namespace	
#endif
