/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.3 $
 *********************************************************************/

#ifndef _RGBCOLORIMAGE_H
#define _RGBCOLORIMAGE_H

#include "Color.h"
#include <Matrix.h>

using namespace util;

namespace prapi
{
	/**
	 * A class that provides methods for easily handling RGB color data.
	 **/
	template <class T=int> class RGBColorImage : public Matrix<RGBColor<T> >
	{
	public:
		/**
		 * Create a square matrix with the given number of rows and
		 * columns.
		 **/
		RGBColorImage(int size=1) : Matrix<RGBColor<T> >(size) {}
		/**
		 * Create an matrix with the given number of rows and columns.
		 **/
		RGBColorImage(int rows, int columns) : Matrix<RGBColor<T> >(rows,columns) {}
		/**
		 * Copy an RGB color image.
		 **/
		RGBColorImage(RGBColorImage& other) : Matrix<RGBColor<T> >(other) {}
		/**
		 * Copy an RGB color image of another content type.
		 **/
		template <class U> RGBColorImage(const Matrix<U>& other) : Matrix<RGBColor<T> >(other) {}
		/**
		 * Copy an RGB color image.
		 **/
		RGBColorImage& operator= (const RGBColorImage& other) { Matrix<RGBColor<T> >::operator=(other); return *this; }
		/**
		 * Set each pixel in this image to the given value.
		 **/
		RGBColorImage& operator= (const RGBColor<T>& value) { Matrix<RGBColor<T> >::operator=(value); return *this; }
		/**
		 * Get the red channel as a monochrome matrix.
		 **/
		Matrix<T> getRed(void) const { return getChannel(0); }
		/**
		 * Get the green channel as a monochrome matrix.
		 **/
		Matrix<T> getGreen(void) const { return getChannel(1); }
		/**
		 * Get the blue channel as a monochrome matrix.
		 **/
		Matrix<T> getBlue(void) const { return getChannel(2); }

		/**
		 * Multiply each multi-channel color pixel with the given
		 * transformation matrix. This method can be used to transform the
		 * whole image to another color space.
		 **/
		template <class U> void transform(Matrix<U>& mat) throw (MatrixException&);
		/**
		 * Create a copy of this image and multiply each multi-channel
		 * color pixel in it with the given transformation matrix. This
		 * method can be used to transform the whole image to another
		 * color space without altering the original image.
		 **/
		template <class U> RGBColorImage getTransform(Matrix<U>& mat) const throw (MatrixException&);

		/**
		 * Copy this image and transform each pixel in the copy to
		 * normalized RGB color space. Each channel is divided by the
		 * intensity of the pixel.
		 **/
		RGBColorImage<double> getNormalizedImage(void) const;
		/**
		 * Get normalized red channel.
		 **/
		Matrix<double> getNormalizedRed(void) const { return getNormalizedChannel(0); }
		/**
		 * Get normalized green channel.
		 **/
		Matrix<double> getNormalizedGreen(void) const { return getNormalizedChannel(1); }
		/**
		 * Get normalized blue channel.
		 **/
		Matrix<double> getNormalizedBlue(void) const { return getNormalizedChannel(2); }

	private:
		Matrix<T> getChannel(int c) const;
		Matrix<double> getNormalizedChannel(int c) const;
	};


	template <class T>
	template <class U> void RGBColorImage<T>::transform(Matrix<U>& mat) throw (MatrixException&)
	{
		for (int i=0;i<_iRows*_iColumns;i++)
			_pData[i] = mat * _pData[i];
	}
	
	template <class T>
	template <class U> RGBColorImage<T> RGBColorImage<T>::getTransform(Matrix<U>& mat) const throw (MatrixException&)
	{
		RGBColorImage<T> result(_iRows, _iColumns);
		for (int i=0;i<_iRows*_iColumns;i++)
			result._pData[i] = mat * _pData[i];
		return result;
	}

	template <class T> Matrix<T> RGBColorImage<T>::getChannel(int c) const
	{
		Matrix<T> result(_iRows,_iColumns);
		T* rData = result.getData();
		for (int i=0;i<_iRows*_iColumns;i++)
			rData[i] = _pData[i][c];
		return result;
	}

	template <class T> RGBColorImage<double> RGBColorImage<T>::getNormalizedImage(void) const
	{
		RGBColorImage<double> result(_iRows,_iColumns);

		Color<double,3>* rData = result.getData();
		for (int i=0;i<_iRows*_iColumns;i++)
			rData[i] = _pData[i].getNormalizedColor();
		return result;
	}

	template <class T> Matrix<double> RGBColorImage<T>::getNormalizedChannel(int c) const
	{
		Matrix<double> result(_iRows,_iColumns);

		double* rData = result.getData();
		for (int i=0;i<_iRows*_iColumns;i++)
			rData[i] = _pData[i].getNormalizedColor()[c];
		return result;
	}
}	
#endif
