/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.3 $
 *********************************************************************/

#ifndef _COLORCUBEEXTRACTOR_H
#define _COLORCUBEEXTRACTOR_H

#include <List.h>
#include <Matrix.h>

#include "../Color.h"
#include "../FeatureExtractor.h"
#include "../Quantizer.h"

using namespace util;

namespace prapi { namespace color {

	/**
	 * ColorCubeExtractor creates a multi-dimensional histogram out of
	 * any color matrix.
	 **/
	template <class T, int comps>
	class ColorCubeExtractor : public FeatureExtractor<int,Matrix< Color <T,comps> > >
	{
	public:
		/**
		 * Create a new ColorCubeExtractor that uses the given Quantizer
		 * for all color channels.
		 * @param quantizer the quantizer
		 **/
		ColorCubeExtractor(Quantizer* quantizer);
		/**
		 * Create a new ColorCubeExtractor that uses different quantizers
		 * for different channels.
		 * @param quantizers an array of Quantizer*'s that are used for the
		 *                   color channels. The array length must equal to
		 *                   the number of color components in the matrices
		 *                   to be processed.
		 **/
		ColorCubeExtractor(Quantizer* quantizers[comps]);
		List<int> getFeatureVector(const Matrix< Color <T,comps> >& mat);

	private:
		Quantizer *quantizers[comps];
	};

	typedef ColorCubeExtractor<int,3> RGBCubeExtractor;
	typedef ColorCubeExtractor<double,3> DoubleRGBCubeExtractor;
	
	
	template <class T, int comps>
	ColorCubeExtractor<T,comps>::ColorCubeExtractor(Quantizer* q)
	{
		for (int i=0;i<comps;i++)
			quantizers[i] = q;
	}

	template <class T, int comps>
	ColorCubeExtractor<T,comps>::ColorCubeExtractor(Quantizer* q[comps])
	{
		for (int i=0;i<comps;i++)
			quantizers[i] = q[i];
	}

	template <class T, int comps>
	List<int> ColorCubeExtractor<T,comps>::getFeatureVector(const Matrix< Color<T,comps> >& mat)
	{
		int levels[comps], multiplier[comps+1];
		multiplier[0] = 1;
		for (int i=0;i<comps;i++)
			{
				levels[i] = quantizers[i]->getLevels();
				multiplier[i+1] = multiplier[i] * levels[i];
			}

		int size = multiplier[comps];
		List<int> result(size);
		result.setLength(size);
		memset(result.getData(),0,sizeof(int)*size);
		
		Color<T,comps>* data = mat.getData();
		for (int i=0;i<mat.getColumns()*mat.getRows();i++,data++)
			{
				int index = 0;
				for (int j=0;j<comps;j++)
					index += quantizers[j]->getBinIndex((*data)[j]) * multiplier[j];
				result[index]++;
			}
		return result;
	}	
}}

#endif
