/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.5 $
 *********************************************************************/

#ifndef _COLORPERCENTILE_H
#define _COLORPERCENTILE_H

#include <List.h>
#include <Math.h>
#include <Matrix.h>
#include "../FeatureExtractor.h"
#include "../Histogram.h"

using namespace util;

namespace prapi { namespace color {

	/**
	 * A feature extractor for calculating color percentiles. The name
	 * may be somewhat misleading because percentiles are actually
	 * calculated from single-channel integer matrices. Typically, however,
	 * the input matrices are color channels extracted from multi-channel
	 * color images, hence the name.<p>
	 *
	 * Color percentiles are calculated from the probability integral of
	 * an image histogram. Percentiles can be calculated for values in
	 * the range [0,1], and the value for percentile x is the index of
	 * the histogram bin at which the probability integral exceeds (or
	 * equals) x.
	 **/
	class ColorPercentile : public FeatureExtractor<int,Matrix<int> >
	{
	public:
		/**
		 * Create a color percentile extractor with the percentiles
		 * intialized to {0.1, 0.2, ..., 0.9, 1.0}.
		 **/
		ColorPercentile();
		/**
		 * Create a color percentile extractor with the given percentiles.
		 *
		 * @param percentiles the color percentiles to calculate
		 **/
		ColorPercentile(const List<double>& percentiles) : _dlstPercentiles(percentiles) {}

		/**
		 * Calculate the currently active color percentiles for an matrix.
		 * The length of the returned vector is equal to the number of
		 * percentiles given in the constructor or set by the
		 * setPercentiles method.
		 **/
		List<int> getFeatureVector(const Matrix<int>& mat) throw (FeatureExtractionException&);

		/**
		 * Set the percentiles to extract.
		 **/
		void setPercentiles(const List<double>& percentiles) { _dlstPercentiles = percentiles; }
		/**
		 * Get the currently active percentiles.
		 **/
		List<double> getPercentiles() const { return _dlstPercentiles; }
		/**
		 * Get the currently active percentiles.
		 **/
		List<double>& percentiles() { return _dlstPercentiles; }
		/**
		 * Get the currently active percentiles.
		 **/
		const List<double>& percentiles() const { return _dlstPercentiles; }

		/**
		 * Calculate a single color percentile.
		 *
		 * @param percentile the percentile to calculate, in the range [0,1]
		 * @param histogram a histogram
		 * @param total the total number of entries in the histogram,
		 *        for optimization.
		 **/
		inline static int getPercentile(double percentile, List<int> histogram, int total=-1);

	private:
		List<double> _dlstPercentiles;
	};
}}
 

#endif
