/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.3 $
 *********************************************************************/

#ifndef _OPPONENTCOLORLBP_H
#define _OPPONENTCOLORLBP_H

#include "CircularLocalSampler.h"
#include "../FeatureExtractor.h"
#include "../Color.h"

namespace prapi { namespace texture {

	/**
	 * OpponentColorLBP is an LBP variant for color images. The
	 * difference between OC-LBP and its gray-scale correspondent is
	 * that in OC-LBP, the center pixel and the neighborhood can reside
	 * on separate color channels, hence the name. There are a total
	 * number of nine possible combinations, as both the center and the
	 * neighbors can be taken from three color channels (R,G,B). For
	 * illustration of the possible neighborhood settings for the case
	 * involving eight neighborhood samples, see the illustration below.
	 * In the figure, the color channel of the neighborhood center is
	 * lower case for clarity.
	 *
	 * <pre>
	 * RRR RRR RRR GGG GGG GGG BBB BBB BBB
	 * RrR RgR RbR GrG GgG GbG BrB BgB BbB
	 * RRR RRR RRR GGG GGG GGG BBB BBB BBB
	 * </pre>
	 **/
	template <class T=int> class OpponentColorLBP :
		public CircularLocalSampler<int, RGBColor<T> >, public FeatureExtractor<int, util::Matrix<RGBColor<T> > >
	{
	public:
		/**
		 * Create a new OpponentColorLBP feature extractor/image transform
		 * with the given center and neighborhood color channels.
		 **/
		OpponentColorLBP(int centerChannel=0, int surroundChannel=0) :
			_ipPatternMap(NULL), _iMaxMappingValue(-1),
			_iCenterChannel(centerChannel), _iSurroundChannel(surroundChannel) {}

		~OpponentColorLBP();

		int getValue(RGBColor<T> center, util::List<RGBColor<T> >& surrounding);
		util::List<int> getFeatureVector(const util::Matrix<RGBColor<T> >& mat) throw (FeatureExtractionException&);

		/**
		 * Set mapping.
		 **/
		void setMapping(int* mapping, int maxVal);
		/**
		 * Set mapping.
		 **/
		void setMapping(LBPMapping::MappingType type);
		
		/**
		 * Set the center and surrounding color channels.
		 **/
		void setChannels(int centerChannel, int surroundChannel)
		{
			_iCenterChannel = centerChannel;
			_iSurroundChannel = surroundChannel;
		}

		/**
		 * Get the color channel of the center pixel.
		 **/
		int getCenterChannel() { return _iCenterChannel; }
		/**
		 * Get the color channel of the neighborhood.
		 **/
		int getSurroundChannel() { return _iSurroundChannel; }

	private:
		int* _ipPatternMap;
		int _iMaxMappingValue;
		int _iCenterChannel, _iSurroundChannel;
	};


	template <class T> OpponentColorLBP<T>::~OpponentColorLBP()
	{
		delete _ipPatternMap;
	}

	template <class T> void OpponentColorLBP<T>::setMapping(int* mapping, int maxVal)
	{
		if (mapping != _ipPatternMap)
			{
				delete _ipPatternMap;
				_ipPatternMap = mapping;

				_iMaxMappingValue = maxVal;
			}
	}
	
	template <class T> void OpponentColorLBP<T>::setMapping(LBPMapping::MappingType type)
	{
		setMapping(LBPMapping::getMapping(_uiSamples, type),
							 LBPMapping::getSize(_uiSamples, type));
	}

	template <class T> int OpponentColorLBP<T>::getValue(RGBColor<T> center, util::List<RGBColor<T> >& surrounding)
	{
		int result = 0, base = 1;
		T threshold = center[_iCenterChannel];
		for (int i=0;i<surrounding.getLength();i++,base<<=1)
			if (surrounding[i][_iSurroundChannel] >= threshold)
				result |= base;
		if (!_ipPatternMap)
			return result;
		else
			return _ipPatternMap[result];
	}

	template <class T> util::List<int> OpponentColorLBP<T>::getFeatureVector(const util::Matrix<RGBColor<T> >& mat)
		throw (FeatureExtractionException&)
	{
		int resultSize = (_ipPatternMap != NULL)? _iMaxMappingValue : (1<<_uiSamples);
		util::List<int> result(resultSize);
		result.setLength(resultSize, 0);
		util::Matrix<int> img(getTransformedImage(mat));
		int* ptr = img.getData();
		for (int i=img.getRows()*img.getColumns();i--;ptr++)
			result[*ptr]++;
		return result;
	}
}}

#endif
