/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.7 $
 *********************************************************************/

#ifndef _SORTEDLIST_H
#define _SORTEDLIST_H

#include "List.h"
#include <values.h>
#include <functional>

namespace util
{
	/**
	 * SortedList is a sorted list of any objects (including primitive
	 * data types). The items added to the list will always maintain
	 * ascending sort order by default. The order may be changed using
	 * the second template parameter. The objects added to the list must
	 * have well-defined copy behavior plus the operators &gt; (or &lt;) and
	 * == defined. They must also provide a default (empty) constructor.
	 * Primitive types can be used as such with no special actions. The
	 * implementation guarantees O(log(N)) complexity for searches. The
	 * complexity for insertions is at worst O(log(N) + N).<p>
	 *
	 * Template parameters:
	 * <ul>
	 * <li>T - the element type</li>
	 * <li>comparator - either greater&lt;T&gt; (ascending,default) or less&lt;T&gt;
	 * (descending).</li>
	 * </ul>
	 **/
	template <class T, class comparator = std::greater<T> > class SortedList : public List<T>
	{
	public:
		/**
		 * Constructs a new sorted list with the given initial capacity
		 * and block size.
		 * @param initialSlots the number of items that can the list can
		 *        initially hold without expanding
		 * @param blockSize the number of slots the list will expand each
		 *        time its capacity is exceeded
		 **/
		SortedList(int initialSlots=16, int blockSize=16);

		/**
		 * Adds an element to the list. The added element must be of some
		 * elementary type or define the operators >, >= and ==. It must
		 * also provide a default (parameterless) constructor.
		 * @param T the element to be added
		 **/
		void addElement(const T& element);

		/**
		 * Get the index of the first occurrence of the given element.
		 **/
		int indexOf(const T& element) const;

		/**
		 * Restrict the maximum size of this list to <i>size</i>. If the
		 * size of the list exceeds this value, the last element will be
		 * discarded.
		 * @param size the maximum size, initialized to MAXINT
		 **/
		void setMaximumSize(int size);
		/**
		 * Get the maximum size of the list.
		 **/
		int getMaximumSize(void) { return _iMaximumSize; }

	private:
		int _iMaximumSize;
		comparator _comparator;
	};

	typedef SortedList<int> SortedIntList;
	typedef SortedList<float> SortedFloatList;
	typedef SortedList<double> SortedDoubleList;

	
	template <class T, class comparator> SortedList<T,comparator>::SortedList(int initialSlots, int blockSize) :
		List<T>(initialSlots, blockSize)
	{
		_iMaximumSize = MAXINT;
	}

	template <class T, class comparator> void SortedList<T,comparator>::addElement(const T& element)
	{
		int start=0,end=_iCurrentItems;
		while(start<end)
			{
				int half = (start+end)>>1;
				if (_comparator(_internalArray[half],element))
					end = half;
				else
					start = half+1;
			}
		insertElementAt(element,start);
		//cerr << "Inserting " << element << " at " << start << ". Items: " << _iCurrentItems << endl;
		if (_iCurrentItems > _iMaximumSize)
			_iCurrentItems = _iMaximumSize;
	}

	template <class T, class comparator> int SortedList<T,comparator>::indexOf(const T& element) const
	{
		int start=0,end=_iCurrentItems;
		while(start<end)
			{
				int half = (start+end)>>1;
				if (!_comparator(element,_internalArray[half]))
					end = half;
				else
					start = half+1;
			}
		return ((end < _iCurrentItems) && (_internalArray[end] == element))? end : -1;
	}

	template <class T, class comparator> void SortedList<T,comparator>::setMaximumSize(int size)
	{
		_iMaximumSize = size;
	}
	
}

#endif
